from astropy.table import Table, vstack
from astropy.coordinates import SkyCoord
from astropy import units as u
from astropy.utils.data import get_pkg_data_filename
import numpy as np
from .. import calc_multi_vmag, calc_vmag


def test_one_vmag():
    find_var_data = get_pkg_data_filename('data/variables.fits')
    var_stars = Table.read(find_var_data)
    find_star_data = get_pkg_data_filename('data/2014-12-29-ey-uma-9.fits')
    star_data = Table.read(find_star_data)
    find_comp_data = get_pkg_data_filename('data/comp_stars.fits')
    comp_stars = Table.read(find_comp_data)
    comp_stars['coords'] = SkyCoord(ra=comp_stars['ra'], dec=comp_stars['dec'], unit='degree')
    vmag, error = calc_vmag(var_stars, star_data, comp_stars, band='Rc', star_data_mag_column='mag_inst_R')
    np.testing.assert_almost_equal(vmag, 14.54827, decimal=5)
    np.testing.assert_almost_equal(error, 0.028685, decimal=5)


def test_multi_vmag():
    find_var_data = get_pkg_data_filename('data/variables.fits')
    var_stars = Table.read(find_var_data)
    find_star_data = get_pkg_data_filename('data/2014-12-29-ey-uma-9.fits')
    star_data = Table.read(find_star_data)
    find_comp_data = get_pkg_data_filename('data/comp_stars.fits')
    comp_stars = Table.read(find_comp_data)
    comp_stars['coords'] = SkyCoord(ra=comp_stars['ra'], dec=comp_stars['dec'], unit='degree')

    vmag, error = calc_vmag(var_stars, star_data, comp_stars, band='Rc', star_data_mag_column='mag_inst_R')
    del var_stars['coords']
    v_data = vstack([var_stars, var_stars])
    v_data['coords'] = SkyCoord(ra=v_data['RAJ2000'],
                                dec=v_data['DEJ2000'], unit='degree')
    v_table = calc_multi_vmag(v_data, star_data, comp_stars, band='Rc', star_data_mag_column='mag_inst_R')

    assert v_table['Mag'][0] == v_table['Mag'][1]
    assert v_table['StDev'][0] == v_table['StDev'][1]



version = 'unknown.dev'
try:
    from importlib_metadata import version as _version, PackageNotFoundError
    version = _version('my-package')
except ImportError:
    from pkg_resources import get_distribution, DistributionNotFound
    try:
        version = get_distribution("my-package").version
    except DistributionNotFound:
        pass
except PackageNotFoundError:
    pass



from pathlib import Path
import re

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table, hstack

import numpy as np

__all__ = ['parse_aij_table', 'ApertureFileAIJ']


class ApertureAIJ:
    """
    Represent the aperture information AstroImageJ saves.
    """
    def __init__(self):
        # Outer annulus radius
        self.rback2 = 41.0

        # Inner annulus radius
        self.rback1 = 27.0

        # Aperture radius
        self.radius = 15.0

        # Defaults for this match the defaults for stellarphot.
        self.removebackstars = True

        # Not sure what this does but stellarphot doesn't use it.
        self.backplane = False

    def __setattr__(self, attr, value):
        floats = ['rback1', 'rback2', 'radius']
        bools = ['removebackstars', 'backplane']
        if attr in floats:
            super().__setattr__(attr, float(value))
        elif attr in bools:
            if isinstance(value, str):
                value = True if value.lower() == 'true' else False

            super().__setattr__(attr, value)

    def __eq__(self, other):
        attributes = ['rback1', 'rback2', 'radius',
                      'removebackstars', 'backplane']
        equal = [getattr(self, attr) == getattr(other, attr)
                 for attr in attributes]

        return all(equal)


class MultiApertureAIJ:
    """
    Represent the multi-aperture information that AstroImageJ saves
    """
    def __init__(self):
        # Default values for these chosen to match AIJ defaults
        # They are not used by stellarphot
        self.naperturesmax = 1000
        self.apfwhmfactor = 1.4
        self.usevarsizeap = False

        # Each attribute below should be list-like, all of the same length
        self.isrefstar = []
        self.centroidstar = []
        self.isalignstar = []

        self.xapertures = []
        self.yapertures = []

        self.absmagapertures = []

        self.raapertures = []
        self.decapertures = []

    def __setattr__(self, name, value):
        floats = ['naperturesmax', 'apfwhmfactor']
        bools = ['usevarsizeap']
        lists = ['isrefstar', 'centroidstar', 'isalignstar', 'xapertures',
                 'yapertures', 'absmagapertures', 'raapertures', 'decapertures']

        list_is_bool = ['isrefstar', 'centroidstar', 'isalignstar']

        if name not in (floats + bools + lists):
            raise AttributeError(f'Attribute {name} does not exist')

        if name in floats:
            value = float(value)
        elif name in bools:
            if isinstance(value, str):
                value = True if value.lower() == 'true' else False
            else:
                value = bool(value)
        elif name in lists:
            if (name in list_is_bool) and (len(value) > 0) and isinstance(value[0], str):
                value = [True if v.lower() == 'true' else False for v in value]
            value = list(value)

        super().__setattr__(name, value)

    def __eq__(self, other):
        simple_attrs = [
            'naperturesmax',
            'apfwhmfactor',
            'usevarsizeap',
            'isrefstar',
            'centroidstar',
            'isalignstar',
        ]

        float_attrs = [
            'xapertures',
            'yapertures',
            'absmagapertures',
            'raapertures',
            'decapertures'
        ]

        simple_eq = [getattr(self, attr) == getattr(other, attr)
                     for attr in simple_attrs]

        float_eq = [np.allclose(getattr(self, attr), getattr(other, attr), equal_nan=True)
                    for attr in float_attrs]

        equal = simple_eq + float_eq

        return all(equal)


class ApertureFileAIJ:
    """
    Represent AstroImageJ aperture file.
    """
    def __init__(self):
        self.aperture = ApertureAIJ()
        self.multiaperture = MultiApertureAIJ()

    def __str__(self):
        lines = []

        top_level_attrib = vars(self)
        for name, attrib in top_level_attrib.items():
            base_attrib = vars(attrib)
            for bname, battrib in base_attrib.items():
                try:
                    value = ','.join([str(v).lower() for v in battrib])
                except TypeError:
                    value = battrib

                if value is True:
                    value = 'true'
                elif value is False:
                    value = 'false'

                lines.append(f'.{name}.{bname}={value}')

        # Add a trailing blank line
        return '\n'.join(lines) + '\n'

    def __eq__(self, other):
        return (self.aperture == other.aperture) and (self.multiaperture == other.multiaperture)

    def write(self, file):
        p = Path(file)
        p.write_text(str(self))

    @classmethod
    def read(cls, file):
        """
        Generate aperture object from file. Happily, each line is basically a path
        to an attribute name followed by a value.
        """

        # Make the instance to return

        aij_aps = cls()

        with open(file, 'r') as f:
            for line in f:
                class_path, value = line.strip().split('=')
                # There is always a leading dot
                _, attr1, attr2 = class_path.split('.')
                obj = getattr(aij_aps, attr1)

                # value is either a single value or a list of values separated by commas
                vals = value.split(',')
                if len(vals) == 1:
                    val_to_set = vals[0]
                else:
                    try:
                        val_to_set = [float(v) for v in vals]
                    except ValueError:
                        val_to_set = list(vals)

                setattr(obj, attr2, val_to_set)

        return aij_aps

    @classmethod
    def from_table(cls, aperture_table,
                   aperture_rad=None, inner_annulus=None, outer_annulus=None,
                   default_absmag=99.999, default_isalign=True,
                   default_centroidstar=True,
                   y_size=4096):
        """
        Create an `stellarphot.io.ApertureFileAIJ` from a stellarphot aperture
        table and info about the aperture sizes.

        Parameters
        ----------

        aperture_table : `astropy.table.Table`
            Table of aperture information.

        aperture_rad : number
            Radius of aperture.

        inner_annulus : number
            Inner radius of annulus.

        outer_annulus : number
            Outer radius of annulus.
        """
        # Create the instance
        apAIJ = cls()

        # Populate aperture properties
        apAIJ.aperture.rback2 = outer_annulus
        apAIJ.aperture.rback1 = inner_annulus
        apAIJ.aperture.radius = aperture_rad

        # Remaining properties of apAIJ.aperture default to the
        # correct values for stellarphot

        # Populate multiaperture properties
        n_apertures = len(aperture_table)
        columns = aperture_table.colnames

        if n_apertures > apAIJ.multiaperture.naperturesmax:
            apAIJ.multiaperture.naperturesmax = n_apertures + 1
        # A boolean column for this would be better, but this will do
        # for now.
        apAIJ.multiaperture.isrefstar = [('comparison' in name.lower())
                           for name in aperture_table['marker name']]

        # These are not currently in the table but that could change...
        if 'centroidstar' not in columns:
            apAIJ.multiaperture.centroidstar = [default_centroidstar] * n_apertures
        else:
            apAIJ.multiaperture.centroidstar = aperture_table['centroidstar']

        if 'isalign' not in columns:
            apAIJ.multiaperture.isalignstar = [default_isalign] * n_apertures
        else:
            apAIJ.multiaperture.isalignstar = aperture_table['isalign']

        apAIJ.multiaperture.xapertures = aperture_table['x']
        apAIJ.multiaperture.yapertures = y_size - aperture_table['y']

        if 'absmag' not in columns:
            apAIJ.multiaperture.absmagapertures = [default_absmag] * n_apertures
        else:
            apAIJ.multiaperture.absmagapertures = aperture_table['absmag']

        apAIJ.multiaperture.raapertures = aperture_table['coord'].ra.degree
        apAIJ.multiaperture.decapertures = aperture_table['coord'].dec.degree

        return apAIJ


def _is_comp(star_coord, comp_table):
    idx, d2d, _ = star_coord.match_to_catalog_sky(comp_table['coord'])
    return 'comparison' in comp_table['marker name'][idx]


def generate_aij_table(table_name, comparison_table):
    info_columns = {
        'date-obs': 'DATE_OBS',
        'airmass': 'AIRMASS',
        'BJD': 'BJD_MOBS',
        'exposure': 'EXPOSURE',
        'filter': 'FILTER',
        'aperture': 'Source_Radius',
        'annulus_inner': 'Sky_Rad(min)',
        'annulus_outer': 'Sky_Rad(max)'
    }
    by_source_columns = {
        'xcenter': 'X(IJ)',
        'ycenter': 'Y(IJ)',
        'aperture_net_flux': 'Source-Sky',
        'aperture_area': 'N_Src_Pixels',
        'noise-aij': 'Source_Error',
        'snr': 'Source_SNR',
        'sky_per_pix_avg': 'Sky/Pixel',
        'annulus_area': 'N_Sky_Pixels',
        'fwhm_x': 'X-Width',
        'fwhm_y': 'Y-Width',
        'width': 'Width',
        'relative_flux': 'rel_flux',
        'relative_flux_error': 'rel_flux_err',
        'relative_flux_snr': 'rel_flux_SNR',
        'comparison counts': 'tot_C_cnts',
        'comparison error': 'tot_C_err'
    }
    by_star = table_name.group_by('star_id')
    base_table = by_star.groups[0][list(info_columns.keys())]
    for star_id, sub_table in zip(by_star.groups.keys, by_star.groups):
        star_co = SkyCoord(ra=sub_table['RA'][0], dec=sub_table['Dec'][0],
                           unit='degree')

        if _is_comp(star_co, comparison_table):
            char = 'C'
        else:
            char = 'T'

        new_table = sub_table[list(by_source_columns.keys())] #  + ['BJD']

        for old_col, new_col in by_source_columns.items():
            new_column_name = new_col + f'_{char}{star_id[0]}'
            new_table.rename_column(old_col, new_column_name)
        # Add individual columns to the existing table instead of hstack
        # Turns out hstack is super slow.
        for new_col in new_table.colnames:
            base_table[new_col] = new_table[new_col]

    for old_col, new_col in info_columns.items():
        base_table.rename_column(old_col, new_col)

    return base_table


def parse_aij_table(table_name):
    """
    Return a list of objects, one for each source, from an AstroImageJ
    photometry table.

    Parameters
    ----------

    table_name : str
        Name of the table.
    """

    # Read in the raw table.
    if table_name.endswith('.csv'):
        # The table may have been edited and changed to csv.
        raw = Table.read(table_name)
    else:
        # The default, though, is tab-separated text with a file extension xls.
        raw = Table.read(table_name, format='ascii.tab')

    # Extract the names of all columns which are not specific to a source.
    # Source columns end with _TX or _CX where X is one or more digits.
    source_column = r'.*_[CT]\d+'
    common_columns = []
    for name in raw.colnames:
        if not re.search(source_column, name):
            if name.strip():
                common_columns.append(name)

    # Get all of the source designations from the names of the net counts
    # columns.
    flux_columns = [name for name in raw.colnames if
                    name.startswith('Source-Sky')]
    source_column_ids = [c.split('_')[1] for c in flux_columns]

    # For the first source, grab all of the column names that are specific to
    # a source. These will be the same for all sources.
    first_source_names = [name for name in raw.colnames if
                          name.endswith(source_column_ids[0])]
    generic_source_columns = [name.rsplit('_', 1)[0] for name in
                              first_source_names]

    # Make a list of star objects. Not sure this is actually better than
    # a list of tables or something simple like that.
    stars = []
    for idx, source in enumerate(source_column_ids):
        specifc_column_names = ['_'.join([g, source]) for g in
                                generic_source_columns]
        all_names = common_columns + specifc_column_names
        my_table = raw[all_names]
        for spec, gen in zip(specifc_column_names, generic_source_columns):
            my_table.rename_column(spec, gen)
        stars.append(Star(my_table, idx + 1))

    return stars


class Star(object):
    def __init__(self, table, id_num):
        self._table = table
        self._table['DEC'].unit = u.degree
        self.id = id_num

    @property
    def airmass(self):
        return self._table['AIRMASS']

    @property
    def counts(self):
        return self._table['Source-Sky']

    @property
    def ra(self):
        return self._table['RA'] / 24 * 360 * u.degree

    @property
    def dec(self):
        return self._table['DEC']

    @property
    def error(self):
        return self._table['Source_Error']

    @property
    def sky_per_pixel(self):
        return self._table['Sky/Pixel']

    @property
    def peak(self):
        return self._table['Peak']

    @property
    def jd_utc_start(self):
        return self._table['JD_UTC']

    @property
    def mjd_start(self):
        return self._table['J.D.-2400000'] - 0.5

    @property
    def exposure(self):
        try:
            return self._table['EXPOSURE']
        except KeyError:
            return self._table['EXPTIME']

    @property
    def magnitude(self):
        return -2.5 * np.log10(self.counts) + 2.5 * np.log10(self.exposure)

    @property
    def snr(self):
        return self.counts / self.error

    @property
    def magnitude_error(self):
        return 1.0857 / self.snr

    @property
    def bjd_tdb(self):
        return self._table['BJD_TDB']



class ObserveACP:

    def __init__(self, name, ra=None, dec=None):
        '''
        imports Skycoord, astropy.time, and datetime

        initialize object with given attributes

        Attributes
        ----------
        name: The name of the observed object
        start_time: time of observation as a string in form 00:00:00
        filters: list including the names of filters to be used in observation. case sensitive.
        exposure: exposure time in integer seconds.
        im_count: the number of images taken in each filter as an int
        repeat: how many times the observation is repeated
        binning: the binning of the CCD. should be kept 1.
        ra: the Right Ascension of the object in decimal degrees
        dec: the Declination of the object in decimal degrees

        Parameters
        ----------
        name: the name of the observed object. must be a string
        ra: the right ascension of the aobject
        dec: the declination of the object

        Returns
        -------
        does note return, None
        '''

        from astropy.coordinates import SkyCoord
        from astropy.time import Time
        import datetime

        # set some defaults
        self.name = name
        self.start_time = str(datetime.datetime.today()).split()[1][:8]
        self.filters = []
        self.exposure = 1
        self.im_count = 1
        self.repeat = 1
        self.binning = 1

        if (ra is None) and (dec is None):

            coords = SkyCoord.from_name(self.name)
            self.ra = coords.ra.deg
            self.dec = coords.dec.deg

        elif (ra is None) or (dec is None):
            raise ValueError("Must give both RA and DEC")

        else:

            self.ra = ra
            self.dec = dec

    def wait_until(self, time):
        '''
        sets a time to start observation

        Parameters
        ----------
        time: string representing the time of observation in the form 01:23:45

        Returns
        -------
        Does not return, None
        '''
        self.start_time = time

    def add_filter(self, filt):

        if filt in self.filters:
            print("Filter", filt, "is already in list!")

        elif isinstance(filt, str):
            # do the string thing
            self.filters.append(filt)

        elif not isinstance(filt, str):
            try:
                # loop over elements in filt
                for i in filt:
                    if i in self.filters:
                        print("Filter", i, "is already in list")

                    else:
                        self.filters.append(i)

            except IndexError:
                raise ValueError('filt must be a non-string iterable')

    def rm_filter(self, filt):

        if filt not in self.filters:
            print("No filter", filt, "in list")

        elif isinstance(filt, str):
            self.filters.remove(filt)

        elif not isinstance(filt, str):
            try:
                for i in filt:
                    if i in self.filters:

                        self.filters.remove(i)

                    else:
                        print("No filter", filt, "in list")
            except IndexError:
                raise ValueError('filt must be a non-string iterable')

    def expose(self, num):
        '''
        set an exposure time to be written to .txt file

        Parameters
        ----------
        num: an integer value of the exposure time.

        Returns
        -------
        Does not return, None
        '''
        self.exposure = num

    def count(self, num):
        '''
        sets how many images of each filter you are taking

        Parameters
        ----------
        num: an integer value of the number of images being taken in one filter

        Returns
        -------
        Does not return, None
        '''
        self.im_count = num

    def repeat(self, num):
        '''
        sets a repeat in the .txt file. determines how many times you repeat the observation.

        Parameters
        ----------
        num: an integer value of the number of repeat observations.

        Returns
        -------
        Does not return, None
        '''
        self.repeat = num

    def write(self, title):
        '''
        writes all class attributes to a text file with a given name. All attributes have default values, but are not ideal. please
        go through each method and set values.

        Parameters
        ----------
        title: A string that defines the name of the text file to be written to.

        Returns
        -------
        Does not return, None.

        '''

        scalar = len(self.filters)

        f = open(title, 'w')

        f.write("#waituntil 1, " + str(self.start_time) + '\n')
        f.write("#repeat " + str(self.repeat) + '\n')

        count = scalar * (str(self.im_count) + ',')
        f.write("#count " + count + '\n')

        f.write("#filter ")
        for i in self.filters:
            f.write(i + ',')
        f.write('\n')

        interval = scalar * (str(self.exposure) + ',')
        f.write("#interval " + interval + '\n')

        binning = scalar * (str(self.binning) + ',')
        f.write("#binning " + binning + '\n')

        RA = str(self.ra)
        DEC = str(self.dec)
        f.write(self.name + '    ' + RA + '    ' + DEC + '\n' + '\n')



import numpy as np
import pytest

from astropy.utils.data import get_pkg_data_filename
from astropy.table import Table
from astropy.stats import gaussian_sigma_to_fwhm
from astropy import units as u

from photutils.datasets import make_gaussian_sources_image, make_noise_image

from stellarphot.source_detection import source_detection, compute_fwhm
from stellarphot.photometry import photutils_stellar_photometry


class FakeImage:
    def __init__(self, noise_dev=1.0):
        self.image_shape = [400, 500]
        data_file = get_pkg_data_filename('data/test_sources.csv')
        self._sources = Table.read(data_file)
        self.mean_noise = self.sources['amplitude'].max() / 100
        self.noise_dev = noise_dev
        self._stars = make_gaussian_sources_image(self.image_shape,
                                                  self.sources)
        self._noise = make_noise_image(self._stars.shape,
                                       mean=self.mean_noise,
                                       stddev=noise_dev)
        self._sources['sky_per_pix_avg'] = noise_dev

    @property
    def sources(self):
        return self._sources

    @property
    def image(self):
        return self._stars + self._noise


@pytest.mark.parametrize('units', [u.pixel, None])
def test_compute_fwhm(units):
    fake_image = FakeImage()
    sources = fake_image.sources
    if units is not None:
        # It turns out having a unit on a column is not the same as
        # things in the column having units. The construct below ensures
        # that the source table values have units.
        # Do not try: sources['x_mean'] = sources['x_mean'] * units
        # Turns out individual values do NOT have units in that case.
        sources['x_mean'] = [v * units for v in sources['x_mean']]
        sources['y_mean'] = [v * units for v in sources['y_mean']]

    fwhm_x, fwhm_y = compute_fwhm(fake_image.image, sources,
                                  x_column='x_mean', y_column='y_mean')


def test_detect_source_number_location():
    """
    Make sure we detect the sources in the input table....
    """
    fake_image = FakeImage()
    sources = fake_image.sources
    print(sources)
    found_sources = source_detection(fake_image.image,
                                     fwhm=2 * sources['x_stddev'].mean(),
                                     threshold=10,
                                     sky_per_pix_avg=sources['sky_per_pix_avg'])
    # Sort by flux so we can reliably match them
    sources.sort('amplitude')
    found_sources.sort('flux')

    # Do we have the right number of sources?
    assert len(sources) == len(found_sources)

    for inp, out in zip(sources, found_sources):
        # Do the positions match?
        np.testing.assert_allclose(out['xcentroid'], inp['x_mean'],
                                   rtol=1e-5, atol=0.05)
        np.testing.assert_allclose(out['ycentroid'], inp['y_mean'],
                                   rtol=1e-5, atol=0.05)
        np.testing.assert_allclose(gaussian_sigma_to_fwhm * (inp['x_stddev'] + inp['y_stddev']) / 2,
                                   out['FWHM'],
                                   rtol=1e-5, atol=0.05)


def test_aperture_photometry_no_outlier_rejection():
    fake_image = FakeImage()
    sources = fake_image.sources
    aperture = sources['aperture'][0]
    found_sources = source_detection(fake_image.image,
                                     fwhm=sources['x_stddev'].mean(),
                                     threshold=10)
    phot = photutils_stellar_photometry(fake_image.image,
                                        found_sources, aperture,
                                        2 * aperture, 3 * aperture,
                                        reject_background_outliers=False)
    phot.sort('aperture_sum')
    sources.sort('amplitude')
    found_sources.sort('flux')

    for inp, out in zip(sources, phot):
        stdev = inp['x_stddev']
        expected_flux = (inp['amplitude'] * 2 * np.pi *
                         stdev**2 *
                         (1 - np.exp(-aperture**2 / (2 * stdev**2))))
        # This expected flux is correct IF there were no noise. With noise, the
        # standard deviation in the sum of the noise within in the aperture is
        # n_pix_in_aperture times the single-pixel standard deviation.
        #
        # We could require that the result be within some reasonable
        # number of those expected variations or we could count up the
        # actual number of background counts at each of the source
        # positions.

        # Here we just check whether any difference is consistent with
        # less than the expected one sigma deviation.
        assert (np.abs(expected_flux - out['net_flux']) <
                np.pi * aperture**2 * fake_image.noise_dev)


@pytest.mark.parametrize('reject', [True, False])
def test_aperture_photometry_with_outlier_rejection(reject):
    """
    Insert some really large pixel values in the annulus and check that
    the photometry is correct when outliers are rejected and is
    incorrect when outliers are not rejected.
    """
    fake_image = FakeImage()
    sources = fake_image.sources
    aperture = sources['aperture'][0]
    image = fake_image.image

    found_sources = source_detection(image,
                                     fwhm=sources['x_stddev'].mean(),
                                     threshold=10)

    inner_annulus = 2 * aperture
    outer_annulus = 3 * aperture
    # Add some large pixel values to the annulus for each source.
    # adding these moves the average pixel value by quite a bit,
    # so we'll only get the correct net flux if these are removed.
    for source in fake_image.sources:
        center_px = (np.int(source['x_mean']), np.int(source['y_mean']))
        begin = center_px[0] + inner_annulus + 1
        end = begin + (outer_annulus - inner_annulus - 1)
        # Yes, x and y are deliberately reversed below.
        image[center_px[1], begin:end] = 100 * fake_image.mean_noise

    phot = photutils_stellar_photometry(image,
                                        found_sources, aperture,
                                        2 * aperture, 3 * aperture,
                                        reject_background_outliers=reject)
    phot.sort('aperture_sum')
    sources.sort('amplitude')
    found_sources.sort('flux')

    for inp, out in zip(sources, phot):
        stdev = inp['x_stddev']
        expected_flux = (inp['amplitude'] * 2 * np.pi *
                         stdev**2 *
                         (1 - np.exp(-aperture**2 / (2 * stdev**2))))
        # This expected flux is correct IF there were no noise. With noise, the
        # standard deviation in the sum of the noise within in the aperture is
        # n_pix_in_aperture times the single-pixel standard deviation.
        #

        expected_deviation = np.pi * aperture**2 * fake_image.noise_dev
        # We could require that the result be within some reasonable
        # number of those expected variations or we could count up the
        # actual number of background counts at each of the source
        # positions.

        # Here we just check whether any difference is consistent with
        # less than the expected one sigma deviation.
        if reject:
            assert (np.abs(expected_flux - out['net_flux']) <
                    expected_deviation)
        else:
            with pytest.raises(AssertionError):
                assert (np.abs(expected_flux - out['net_flux']) <
                        expected_deviation)



import numpy as np


def plot_many_factors(photometry, low, high, shift, scale):
    airmass = photometry['airmass'] / np.mean(photometry['airmass'])
    x = photometry['xcenter'] / np.mean(photometry['xcenter'])
    y = photometry['ycenter'] / np.mean(photometry['ycenter'])
    comp_counts = photometry['comparison counts'] / np.mean(photometry['comparison counts'])
    sky_per_pix = photometry['sky_per_pix_avg'] / np.mean(photometry['sky_per_pix_avg'])
    width = photometry['width'] / np.mean(photometry['width'])

    scale_airmass = scale_and_shift(airmass, scale, 0.75 * shift, pos=False)
    scale_x = scale_and_shift(x, scale, shift, pos=True)
    scale_y = scale_and_shift(y, scale, shift, pos=True)
    scale_sky_pix = scale_and_shift(sky_per_pix, scale, shift, pos=True)
    scale_counts = scale_and_shift(comp_counts, scale, shift, pos=True)
    scale_width = scale_and_shift(width, scale, shift, pos=True)

    grid_y_ticks = np.arange(low, high, 0.02)


def bin_data(data_set, num=3, error_set=None):
    binned_set = []
    error = []
    for i in range(0, len(data_set), num):
        binned_set.append(data_set[i:i+num].mean())
        if error_set is not None:
            error_bin = error_set[i:i+num]**2
            error.append(error_bin.sum()/num)
    return np.array(binned_set), np.array(error)


def scale_and_shift(data_set, scale, shift, pos=True):
    if not pos:
        data_set = 1 - scale * (data_set - data_set.min()) / (data_set.max() - data_set.min())

    else:
        data_set = 1 + scale * (data_set - data_set.min()) / (data_set.max() - data_set.min())
    data_set += shift

    return data_set



import warnings

import numpy as np

from astropy.modeling.models import Polynomial1D, Gaussian1D, custom_model
from astropy.modeling.fitting import (LevMarLSQFitter,
                                      _validate_model,
                                      _convert_input)

# Functions blow changed from private to public in astropy 5
try:
    from astropy.modeling.fitting import (
        fitter_to_model_params,
        model_to_fit_params,
    )
except ImportError:
    from astropy.modeling.fitting import (
        _fitter_to_model_params as fitter_to_model_params,
        _model_to_fit_params as model_to_fit_params,
    )

from astropy.utils.exceptions import AstropyUserWarning

try:
    import batman
except ImportError:
    ImportError('You must install the batman exoplanet package. Try:\n'
                'conda install batman-package\n'
                'or\n'
                'pip install batman-package')


class VariableArgsFitter(LevMarLSQFitter):
    """
    Allow fitting of functions with arbitrary number of positional
    parameters.
    """
    def __init__(self):
        super().__init__()

    # This is a straight copy-paste from the LevMarLSQFitter __call__.
    # The only modification is to allow any number of arguments.
    def __call__(self, model, *args, weights=None,
                 maxiter=100, acc=1e-7,
                 epsilon=1.4901161193847656e-08, estimate_jacobian=False):
        from scipy import optimize

        model_copy = _validate_model(model, self.supported_constraints)
        farg = (model_copy, weights, ) + args
        if model_copy.fit_deriv is None or estimate_jacobian:
            dfunc = None
        else:
            dfunc = self._wrap_deriv
        init_values = model_to_fit_params(model_copy)
        fitparams, cov_x, dinfo, mess, ierr = optimize.leastsq(
            self.objective_function, init_values, args=farg, Dfun=dfunc,
            col_deriv=model_copy.col_fit_deriv, maxfev=maxiter, epsfcn=epsilon,
            xtol=acc, full_output=True)
        fitter_to_model_params(model_copy, fitparams)
        self.fit_info.update(dinfo)
        self.fit_info['cov_x'] = cov_x
        self.fit_info['message'] = mess
        self.fit_info['ierr'] = ierr
        if ierr not in [1, 2, 3, 4]:
            warnings.warn("The fit may be unsuccessful; check "
                          "fit_info['message'] for more information.",
                          AstropyUserWarning)

        # now try to compute the true covariance matrix
        if (len(args[-1]) > len(init_values)) and cov_x is not None:
            sum_sqrs = np.sum(self.objective_function(fitparams, *farg)**2)
            dof = len(args[-1]) - len(init_values)
            self.fit_info['param_cov'] = cov_x * sum_sqrs / dof
        else:
            self.fit_info['param_cov'] = None

        return model_copy


class TransitModelFit:
    def __init__(self, batman_params=None):
        self._batman_params = batman.TransitParams()
        self._set_default_batman_params()
        self._times = None
        self._airmass = None
        self._spp = None
        self._width = None
        self._fitter = VariableArgsFitter()
        self._model = None
        self._batman_mod_for_fit = None
        self.weights = None
        self._detrend_parameters = set()
        self._all_detrend_params = ['airmass', 'width', 'spp']

    def _check_consistent_lengths(self, proposed_value):
        """
        Check that the proposed value has a length consistent with the
        other independent variables. Consistent means same length as others
        or the others are ``None``.
        """
        if proposed_value is None:
            return True

        new_length = len(proposed_value)
        for independent_var in [self._times, self._airmass,
                                self._spp, self._width]:
            if independent_var is None:
                continue
            elif len(independent_var) != new_length:
                return False
        else:
            # All the lengths were good
            return True

    @property
    def times(self):
        return self._times

    @times.setter
    def times(self, value):
        if not self._check_consistent_lengths(value):
            raise ValueError('Length of times not consistent with '
                             'other independent variables.')
        self._times = value

        try:
            if self._batman_mod_for_fit is None:
                self._set_up_batman_model_for_fitting()
        except ValueError:
            pass

    @property
    def airmass(self):
        return self._airmass

    @airmass.setter
    def airmass(self, value):
        if not self._check_consistent_lengths(value):
            raise ValueError('Length of airmass not consistent with '
                             'other independent variables.')
        self._airmass = value

        if value is not None:
            self._detrend_parameters.add('airmass')
        else:
            self._detrend_parameters.discard('airmass')

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, value):
        if not self._check_consistent_lengths(value):
            raise ValueError('Length of width not consistent with '
                             'other independent variables.')
        self._width = value

        if value is not None:
            self._detrend_parameters.add('width')
        else:
            self._detrend_parameters.discard('width')

    @property
    def spp(self):
        return self._spp

    @spp.setter
    def spp(self, value):
        if not self._check_consistent_lengths(value):
            raise ValueError('Length of spp not consistent with '
                             'other independent variables.')
        self._spp = value

        if value is not None:
            self._detrend_parameters.add('spp')
        else:
            self._detrend_parameters.discard('spp')

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        if not self._check_consistent_lengths(value):
            raise ValueError('Length of data not consistent with '
                             'independent variables.')
        self._data = value

    @property
    def model(self):
        return self._model

    def _set_default_batman_params(self):
        """
        Initialize batman parameters to some not nonsensical values. These
        will need to be adjusted before fitting.
        """
        # time of inferior conjunction --> THE MIDPOINT of the transit
        self._batman_params.t0 = 59028.723

        # orbital period
        self._batman_params.per = 5.72

        # planet radius (in units of stellar radii)  --> DETERMINES DEPTH (affects duration)
        self._batman_params.rp = 0.035

        # semi-major axis (in units of stellar radii) --> DETERMINES DURATION
        self._batman_params.a = 12.2

        # orbital inclination (in degrees)
        self._batman_params.inc = 90.

        # eccentricity
        self._batman_params.ecc = 0.

        # longitude of periastron (in degrees)
        self._batman_params.w = 90.

        # limb darkening model
        self._batman_params.limb_dark = "quadratic"

        # limb darkening coefficients [u1, u2]
        self._batman_params.u = [0.3, 0.3]

    def _set_up_batman_model_for_fitting(self):
        if self._times is None:
            raise ValueError('times need to be set before setting up '
                             'transit model.')
        self._batman_mod_for_fit = \
            batman.TransitModel(self._batman_params,
                                self.times)

    def _setup_transit_model(self):
        """
        Creates transit astropy model with exoplanet flux and other trends.
        Called when the necessary information is present for setting up the
        model.
        """

        def transit_model_with_trends(time, airmass, width, sky_per_pix,
                                      t0=0.0, period=1.0, rp=0.1, a=10.0,
                                      inclination=90.0, eccentricity=0.0,
                                      limb_u1=0.3, limb_u2=0.3,
                                      airmass_trend=0.0, width_trend=0.0,
                                      spp_trend=0.0):
            self._batman_params.t0 = t0
            self._batman_params.per = period
            self._batman_params.rp = rp
            self._batman_params.a = a
            self._batman_params.inc = inclination
            self._batman_params.ecc = eccentricity
            self._batman_params.u = [limb_u1, limb_u2]

            flux = self._batman_mod_for_fit.light_curve(self._batman_params)
            flux += airmass_trend * (airmass)
            flux += width_trend * width
            flux += spp_trend * sky_per_pix
            return flux

        ModelClass = custom_model(transit_model_with_trends)

        self._model = ModelClass()

        # Set up some defaults for what is fixed
        self._model.period.fixed = True
        self._model.eccentricity.fixed = True
        self._model.limb_u1.fixed = True
        self._model.limb_u2.fixed = True

        # Set some default bounds for inclination, in degrees.
        self._model.inclination.bounds = (80, 90)

        # Planet radius cannot be too small or too big
        self._model.rp.bounds = (0.01, 0.5)

    def setup_model(self, t0=0, depth=0, duration=0,
                    period=1, inclination=90,
                    airmass_trend=0.0,
                    width_trend=0.0,
                    spp_trend=0.0):
        """
        Configure a transit model for fitting. The ``duration`` and ``depth``
        are used to estimate underlying fit parameters; they are not
        themselves fit parameters.

        Parameters
        ----------
        t0 : float
            Time of the center of the transit. Can be in any units
            but should be consistent with the units for the ``period``
            and for the times used for fitting.

        depth : float
            Depth of the transit in parts per thousand.

        duration : float
            Duration of the transit,in the same units as ``t0`` and
            ``period``.
        period : float
            Period of the planet. Should be in the same units as ``t0``
            and times used for fitting.

        inclination : float
            Inclination of the orbit, in degrees.
        """
        self._setup_transit_model()

        # rp is related to depth in a straighforward way
        self._model.rp.value = self._batman_params.rp = np.sqrt(depth / 1000)

        # The estimate below assumes a circular orbit and inclination of
        # 90 degrees (edge on). This should be fine as a starting point
        # for the fit.
        #
        # See Kipping, eq. 16 at
        # https://doi.org/10.1111/j.1365-2966.2010.16894.x
        estimated_a = 1 / np.sin(duration * np.pi / period)
        self._model.a.value = self._batman_params.a = estimated_a

        self._model.period.value = self._batman_params.per = period
        self._model.inclination.value = self._batman_params.inc = inclination
        self._model.t0.value = self._batman_params.t0 = t0
        self._model.airmass_trend.value = airmass_trend
        self._model.width_trend.value = width_trend
        self._model.spp_trend.value = spp_trend

        try:
            if self._batman_mod_for_fit is None:
                self._set_up_batman_model_for_fitting()
        except ValueError:
            pass

    def fit(self):
        """
        Perform a fit and update the model with best-fit values.
        """
        # Maybe do some correctness check of the model before starting.
        if self.times is None:
            raise ValueError('The times must be set before trying '
                             'to fit.')
        if self._model is None:
            raise ValueError('Run setup_model() before trying to fit.')

        if self._batman_mod_for_fit is None:
            self._set_up_batman_model_for_fitting()

        # Check whether any data bits are None and fix those
        # parameters.
        original_values = {}

        if self.spp is None:
            original_values['spp_trend'] = \
                self.model.spp_trend.fixed
            self.model.spp_trend = 0
            self.model.spp_trend.fixed = True
            spp = np.zeros_like(self.times)
        else:
            spp = self.spp

        if self.airmass is None:
            original_values['airmass_trend'] = \
                self.model.airmass_trend.fixed
            self.model.airmass_trend = 0
            self.model.airmass_trend.fixed = True
            airmass = np.zeros_like(self.times)
        else:
            airmass = self.airmass

        if self.width is None:
            original_values['width_trend'] = \
                self.model.width_trend.fixed
            self.model.width_trend = 0
            self.model.width_trend.fixed = True
            width = np.zeros_like(self.times)
        else:
            width = self.width

        # Do the fitting

        new_model = self._fitter(self.model,
                                 self.times,
                                 airmass,
                                 width,
                                 spp,
                                 self.data,
                                 weights=self.weights)

        # Update the model (might not be necessary but can't hurt)
        self._model = new_model
        self._actual_fixed_params = self._model.fixed

        # reset parameters to their original values
        for k, v in original_values.items():
            param = getattr(self.model, k)
            param.fixed = v

    def _detrend(self, model, detrend_by):
        if detrend_by == 'all':
            detrend_by = [p for p in self._all_detrend_params
                          if p in self._detrend_parameters]
        elif isinstance(detrend_by, str):
            detrend_by = [detrend_by]

        detrended = model.copy()
        for trend in detrend_by:
            detrended = detrended - (getattr(self.model, f'{trend}_trend')
                                     * getattr(self, trend))

        return detrended

    def data_light_curve(self, data=None, detrend_by=None):
        data = data if data is not None else self.data

        if detrend_by is not None:
            data = self._detrend(data, detrend_by)

        return data

    def model_light_curve(self, at_times=None, detrend_by=None):
        """
        Calculate the light curve corresponding to the model, optionally
        detrended by one or more parameters.
        """
        zeros = np.zeros_like(self.times)
        airmass = self.airmass if self.airmass is not None else zeros
        width = self.width if self.width is not None else zeros
        spp = self.spp if self.spp is not None else zeros

        if at_times is not None:
            # temporarily reset the batman model to the new times,
            # then restore it.
            original_model = self._batman_mod_for_fit

            self._batman_mod_for_fit = batman.TransitModel(self._batman_params,
                                                           at_times)
            model = self.model(at_times, airmass, width, spp)
            self._batman_mod_for_fit = original_model
        else:
            model = self.model(self.times, airmass, width, spp)

        if detrend_by is not None:
            model = self._detrend(model, detrend_by)

        return model

    @property
    def n_fit_parameters(self):
        return sum(not v for k, v in self._actual_fixed_params.items())

    @property
    def BIC(self):
        residual = self.data - self.model_light_curve()
        chi_sq = ((residual * self.weights)**2).sum()
        BIC = chi_sq + self.n_fit_parameters * np.log(len(self.data))
        return BIC

# example use

# self.model.eccentricity.fixed = True
# self.model.spp_trend = 0
# self.model.spp_trend.fixed = True

# self.model.t0 = ...
# self.model.rp = ...
# self.fit(time, airmass, width, sky_per_pix)
# self.model.rp
# self.model.rp.error

# my_model = TransitModelFit()
# my_model.times = ...
# my_model.airmass = ...

# my_model.setup_model()
# my_model.model.t0 = ...
# my_model.model.rp = ...
# my_model.model.period = 123
# my_model.model.period.fixed = True

# my_model.data = ...

# fit_model = my_model.fit()

# my_model.model_light_curve(detrend_by=['airmass', 'spp'])

# my_model.model_light_curve(at_times=np.linspace(start - 0.1, end + 0.1, num=1000),
#                            detrend_by=['airmass'])

# my_model.model



import pytest
import numpy as np

from stellarphot.photometry import calculate_noise
from stellarphot.core import Camera

GAINS = [1.0, 1.5, 2.0]


def test_calc_noise_defaults():
    # If we put in nothing we should get zero back
    assert calculate_noise() == 0


@pytest.mark.parametrize('aperture_area', [5, 20])
@pytest.mark.parametrize('gain', GAINS)
def test_calc_noise_source_only(gain, aperture_area):
    # If the only source of noise is Poisson error in the source
    # then the noise should be the square root of the counts.
    counts = 100
    expected = np.sqrt(gain * counts)

    np.testing.assert_allclose(calculate_noise(gain=gain,
                                               flux=counts,
                                               aperture_area=aperture_area),
                               expected)


@pytest.mark.parametrize('aperture_area', [5, 20])
@pytest.mark.parametrize('gain', GAINS)
def test_calc_noise_dark_only(gain, aperture_area):
    # Gain should not affect this one. Dark current needs a couple other things,
    # but this is basically Poisson error.
    dark_current = 10
    exposure = 20
    expected = np.sqrt(dark_current * aperture_area * exposure)

    np.testing.assert_allclose(calculate_noise(gain=gain,
                                               dark_current_per_sec=dark_current,
                                               aperture_area=aperture_area,
                                               exposure=exposure),
                               expected)


@pytest.mark.parametrize('aperture_area', [5, 20])
@pytest.mark.parametrize('gain', GAINS)
def test_calc_read_noise_only(gain, aperture_area):
    # The read noise per pixel IS the noise. The only multiplier is
    # the number of pixels.
    read_noise = 10
    expected = np.sqrt(aperture_area * read_noise**2)

    np.testing.assert_allclose(calculate_noise(gain=gain,
                                               read_noise=read_noise,
                                               aperture_area=aperture_area),
                               expected)


@pytest.mark.parametrize('aperture_area', [5, 20])
@pytest.mark.parametrize('gain', GAINS)
def test_calc_sky_only(gain, aperture_area):
    # The sky noise per pixel is the poisson and per pixel.
    sky = 10
    expected = np.sqrt(gain * aperture_area * sky)

    np.testing.assert_allclose(calculate_noise(gain=gain,
                                               aperture_area=aperture_area,
                                               sky_per_pix=sky),
                               expected)


def test_annulus_area_term():
    # Test that noise is correct with an annulus
    aperture_area = 20

    # Annulus is typically quite a bit larger than aperture.
    annulus_area = 10 * aperture_area
    gain = 1.5
    sky = 10
    expected = np.sqrt(gain * aperture_area *
                       (1 + aperture_area / annulus_area) * sky)
    np.testing.assert_allclose(calculate_noise(gain=gain,
                                               aperture_area=aperture_area,
                                               annulus_area=annulus_area,
                                               sky_per_pix=sky),
                               expected)


@pytest.mark.parametrize('digit,expected',
                         ((False, 89.078616), (True, 89.10182)))
def test_calc_noise_messy_case(digit, expected):
    # Do a single test where all the parameters are set and compare with
    # what a calculator gave.
    counts = 1000

    aperture_area = 20
    annulus_area = 10 * aperture_area

    gain = 1.5
    sky = 15
    dark_current = 7
    exposure = 18
    read_noise = 12

    np.testing.assert_allclose(
        calculate_noise(flux=counts,
                        gain=gain,
                        dark_current_per_sec=dark_current,
                        read_noise=read_noise,
                        sky_per_pix=sky,
                        exposure=exposure,
                        aperture_area=aperture_area,
                        annulus_area=annulus_area,
                        include_digitization=digit),
        expected
    )



import numpy as np

from astropy.nddata import CCDData

from ..coordinates import convert_pixel_wcs
from .make_wcs import make_wcs


def test_coord_conversion():
    wcs = make_wcs()
    ccd = CCDData(np.ones([10, 10]), wcs=wcs, unit='adu')
    # Pixel values below should give back ra, dec values for crval
    ra, dec = convert_pixel_wcs(ccd, 4, 4)
    print(ra, dec)
    np.testing.assert_almost_equal(ra, wcs.wcs.crval[0])
    np.testing.assert_almost_equal(dec, wcs.wcs.crval[1])
    # These pixel values should be one larger than the crval since cdelt is 1
    ra, dec = convert_pixel_wcs(ccd, 5, 5)
    np.testing.assert_almost_equal(ra, wcs.wcs.crval[0] + 1, decimal=2)
    np.testing.assert_almost_equal(dec, wcs.wcs.crval[1] + 1, decimal=2)

    # Now transform back...
    x, y = convert_pixel_wcs(ccd, ra, dec, is_pix=False)
    np.testing.assert_almost_equal(x, 5)
    np.testing.assert_almost_equal(y, 5)



import functools
from pathlib import Path

import pandas

import ipywidgets as ipw

import numpy as np

from astropy.table import Table
from astropy.coordinates import SkyCoord
from astropy import units as u
from astropy.nddata import CCDData
from astropy import units as u

try:
    from astrowidgets import ImageWidget
except ImportError:
    from astrowidgets.ginga import ImageWidget

from stellarphot.differential_photometry import *
from stellarphot.photometry import *
from stellarphot.visualization.seeing_profile_functions import set_keybindings


__all__ = ['read_file', 'set_up', 'match', 'mag_scale',
           'in_field', 'make_markers', 'wrap']


def read_file(radec_file):
    """
    Read an AIJ radec file with target and/or comp positions

    Parameters
    ----------

    radec_file : str
        Name of the file

    Returns
    -------

    `astropy.table.Table`
        Table with target information, including a
        `astropy.coordinates.SkyCoord` column.
    """
    df = pandas.read_csv(radec_file, names=['RA', 'Dec', 'a', 'b', 'Mag'])
    target_table = Table.from_pandas(df)
    ra = target_table['RA']
    dec = target_table['Dec']
    target_table['coords'] = SkyCoord(ra=ra, dec=dec, unit=(u.hour, u.degree))
    return target_table


def set_up(sample_image_for_finding_stars,
           directory_with_images='.'):
    """
    Find known variable in the field of view and read sample image.

    Parameters
    ----------

    sample_image_for_finding_stars : str
        Name or URL of a fits image of the field of view.

    directory_with_images : str, optional
        Folder in which the image is located. Ignored if the sample image
        is a URL.
    """
    if sample_image_for_finding_stars.startswith('http'):
        path = sample_image_for_finding_stars
    else:
        path = Path(directory_with_images) / sample_image_for_finding_stars

    ccd = CCDData.read(path)
    try:
        vsx = find_known_variables(ccd)
    except RuntimeError:
        vsx = []
    else:
        ra = vsx['RAJ2000']
        dec = vsx['DEJ2000']
        vsx['coords'] = SkyCoord(ra=ra, dec=dec, unit=(u.hour, u.degree))

    return ccd, vsx


def match(CCD, RD, vsx):
    """
    Find APASS stars in FOV and matches APASS stars to VSX and
    APASS to input targets.
    """
    apass, apass_in_bright = find_apass_stars(
        CCD)
    ra = apass['RAJ2000']
    dec = apass['DEJ2000']
    apass['coords'] = SkyCoord(ra=ra, dec=dec, unit=(u.hour, u.degree))
    apass_coord = apass['coords']

    if vsx:
        v_index, v_angle, v_dist = \
            apass_coord.match_to_catalog_sky(vsx['coords'])
    else:
        v_angle = []
    if RD:
        RD_index, RD_angle, RD_dist = \
            apass_coord.match_to_catalog_sky(RD['coords'])
    else:
        RD_angle = []
    return apass, v_angle, RD_angle


def mag_scale(cmag, apass, v_angle, RD_angle,
              brighter_dmag=0.44, dimmer_dmag=0.75):
    """
    Select comparison stars that are 1) not close the VSX stars or to other
    target stars and 2) fall within a particular magnitude range.


    """
    high_mag = apass['r_mag'] < cmag + dimmer_dmag
    low_mag = apass['r_mag'] > cmag - brighter_dmag
    if v_angle:
        good_v_angle = v_angle > 1.0 * u.arcsec
    else:
        good_v_angle = True

    if RD_angle:
        good_RD_angle = RD_angle > 1.0 * u.arcsec
    else:
        good_RD_angle = True

    good_stars = high_mag & low_mag & good_RD_angle & good_v_angle
    good_apass = apass[good_stars]
    apass_good_coord = good_apass['coords']
    return apass_good_coord, good_stars


def in_field(apass_good_coord, ccd, apass, good_stars):
    """
    Return apass stars in the field of view
    """
    apassx, apassy = ccd.wcs.all_world2pix(
        apass_good_coord.ra, apass_good_coord.dec, 0)
    ccdx, ccdy = ccd.shape

    xin = (apassx < ccdx) & (0 < apassx)
    yin = (apassy < ccdy) & (0 < apassy)
    xy_in = xin & yin
    apass_good_coord[xy_in]
    nt = apass[good_stars]
    ent = nt[xy_in]
    return ent


def make_markers(iw, ccd, RD, vsx, ent,
                 name_or_coord=None):
    """
    Add markers for APASS, TESS targets, VSX.
    Also center on object/coordinate.
    """
    iw.load_nddata(ccd)
    iw.zoom_level = 'fit'
    try:
        iw.reset_markers()
    except AttributeError:
        iw.remove_all_markers()

    if RD:
        iw.marker = {'type': 'circle', 'color': 'green', 'radius': 10}
        iw.add_markers(RD, skycoord_colname='coords',
                       use_skycoord=True, marker_name='TESS Targets')

    if name_or_coord is not None:
        if isinstance(name_or_coord, str):
            iw.center_on(SkyCoord.from_name(name_or_coord))
        else:
            iw.center_on(name_or_coord)

    if vsx:
        iw.marker = {'type': 'circle', 'color': 'blue', 'radius': 10}
        iw.add_markers(vsx, skycoord_colname='coords',
                       use_skycoord=True, marker_name='VSX')

    iw.marker = {'type': 'circle', 'color': 'red', 'radius': 10}
    iw.add_markers(ent, skycoord_colname='coords',
                   use_skycoord=True, marker_name='APASS comparison')
    iw.marker = {'type': 'cross', 'color': 'red', 'radius': 6}


def wrap(imagewidget, outputwidget):
    """
    Make the bits that let you click to select/deselect comparisons
    """
    def cb(viewer, event, data_x, data_y):
        i = imagewidget._viewer.get_image()

        try:
            imagewidget.next_elim += 1
        except AttributeError:
            imagewidget.next_elim = 1
        pad = 15
        x = int(np.floor(event.data_x))
        y = int(np.floor(event.data_y))
        ra, dec = i.wcs.wcs.all_pix2world(event.data_x, event.data_y, 0)
        out_skycoord = SkyCoord(ra=ra, dec=dec, unit=(u.degree, u.degree))

        try:
            all_table = imagewidget.get_markers(marker_name='all')
        except AttributeError:
            all_table = imagewidget.get_all_markers()

        with outputwidget:
            index, d2d, d3d = out_skycoord.match_to_catalog_sky(
                all_table['coord'])
            if d2d < 10 * u.arcsec:
                mouse = all_table['coord'][index].separation(
                    all_table['coord'])
                rat = mouse < 1 * u.arcsec
                elims = [name for name in all_table['marker name']
                         [rat] if name.startswith('elim')]
                if not elims:
                    imagewidget.add_markers(all_table[rat], skycoord_colname='coord', use_skycoord=True,
                                            marker_name=f'elim{imagewidget.next_elim}')
                else:
                    for elim in elims:
                        try:
                            imagewidget.remove_markers_by_name(marker_name=elim)
                        except AttributeError:
                            imagewidget.remove_markers(marker_name=elim)

            else:
                print('sorry try again')
                imagewidget._viewer.onscreen_message('Click closer to a star')
            print(all_table['marker name'][index])
            print(x, y, ra, dec, out_skycoord)
    return cb


class ComparisonViewer:
    def __init__(self,
                 image,
                 directory='.',
                 target_mag=10,
                 bright_mag_limit=8,
                 dim_mag_limit=17,
                 targets_from_file=None,
                 object_coordinate=None,
                 aperture_output_file=None):

        self._label_name = 'labels'
        self._circle_name = 'target circle'
        self.ccd, self.vsx = \
            set_up(image,
                   directory_with_images=directory
                   )

        apass, vsx_apass_angle, targets_apass_angle = match(self.ccd,
                                                            targets_from_file,
                                                            self.vsx)

        apass_good_coord, good_stars = mag_scale(target_mag, apass, vsx_apass_angle,
                                                 targets_apass_angle,
                                                 brighter_dmag=target_mag - bright_mag_limit,
                                                 dimmer_dmag=dim_mag_limit - target_mag)

        apass_comps = in_field(apass_good_coord, self.ccd, apass, good_stars)

        self.box, self.iw = self._viewer()

        make_markers(self.iw, self.ccd, targets_from_file, self.vsx, apass_comps,
                     name_or_coord=object_coordinate)

        self.target_coord = object_coordinate
        self.aperture_output_file = aperture_output_file

        self._make_observers()

    @property
    def variables(self):
        comp_table = self.generate_table()
        new_vsx_mark = comp_table['marker name'] == 'VSX'
        idx, _, _ = comp_table['coord'][new_vsx_mark].match_to_catalog_sky(self.vsx['coords'])
        our_vsx = self.vsx[idx]
        our_vsx['star_id'] = comp_table['star_id'][new_vsx_mark]
        return our_vsx

    def _make_observers(self):
        self._show_labels_button.observe(self._show_label_button_handler,
                                         names='value')
        self._save_var_info.on_click(self._save_variables_to_file)
        self._save_aperture_file.on_click(self._save_aperture_to_file)

    def _save_variables_to_file(self, button=None, filename=''):
        if not filename:
            filename = 'variables.csv'
        self.variables.write(filename)

    def _show_label_button_handler(self, change):
        value = change['new']
        if value:
            # Showing labels can take a bit, disable while in progress
            self._show_labels_button.disabled = True
            self.show_labels()
            self._show_labels_button.disabled = False
        else:
            self.remove_labels()
        self._show_labels_button.description = self._show_labels_button.descriptions[value]

    def _save_aperture_to_file(self, button=None, filename=''):
        if not filename:
            filename = self.aperture_output_file
        print(filename)
        self.generate_table().write(filename)

    def _make_control_bar(self):
        self._show_labels_button = ipw.ToggleButton(description='Click to show labels')
        self._show_labels_button.descriptions = {
            True: 'Click to hide labels',
            False: 'Click to show labels'
        }

        self._save_var_info = ipw.Button(description='Save variable info')

        self._save_aperture_file = ipw.Button(description='Save aperture file')

        controls = ipw.HBox(
            children=[
                self._show_labels_button,
                self._save_var_info,
                self._save_aperture_file
            ]
        )

        return controls

    def _viewer(self):
        header = ipw.HTML(value="""
        <h2>Click and drag or use arrow keys to pan, use +/- keys to zoom</h2>
        <h3>Shift-left click (or Crtl-left click)to exclude star as target
        or comp. Click again to include.</h3>
        """)

        legend = ipw.HTML(value="""
        <h3>Green circles -- Gaia stars within 2.5 arcmin of target</h3>
        <h3>Red circles -- APASS stars within 1 mag of target</h3>
        <h3>Blue circles -- VSX variables</h3>
        <h3>Red × -- Exclude as target or comp</h3>
        """)

        iw = ImageWidget()
        out = ipw.Output()
        set_keybindings(iw)
        bind_map = iw._viewer.get_bindmap()
        gvc = iw._viewer.get_canvas()
        bind_map.map_event(None, ('shift',), 'ms_left', 'cursor')
        gvc.add_callback('cursor-down', wrap(iw, out))

        controls = self._make_control_bar()
        box = ipw.VBox()
        inner_box = ipw.HBox()
        inner_box.children = [iw, legend]
        box.children = [header, inner_box, controls]

        return box, iw

    def generate_table(self):
        try:
            all_table = self.iw.get_all_markers()
        except AttributeError:
            all_table = self.iw.get_markers(marker_name='all')

        elims = np.array([name.startswith('elim')
                         for name in all_table['marker name']])
        elim_table = all_table[elims]
        comp_table = all_table[~elims]

        index, d2d, d3d = elim_table['coord'].match_to_catalog_sky(comp_table['coord'])
        comp_table.remove_rows(index)

        # Calculate how far each is from target
        comp_table['separation'] = self.target_coord.separation(comp_table['coord'])

        # Add dummy column for sorting in the order we want
        comp_table['sort'] = np.zeros(len(comp_table))

        # Set sort order
        apass_mark = comp_table['marker name'] == 'APASS comparison'
        vsx_mark = comp_table['marker name'] == 'VSX'
        tess_mark = ((comp_table['marker name'] == 'TESS Targets') |
                (comp_table['separation'] < 0.3 * u.arcsec))


        comp_table['sort'][apass_mark] = 2
        comp_table['sort'][vsx_mark] = 1
        comp_table['sort'][tess_mark] = 0

        # Sort the table
        comp_table.sort(['sort', 'separation'])

        # Assign the IDs
        comp_table['star_id'] = range(1, len(comp_table) + 1)

        return comp_table

    def show_labels(self):
        plot_names = []
        comp_table = self.generate_table()

        for star in comp_table:
            star_id = star['star_id']
            if star['marker name'] == 'TESS Targets':
                label = f'T{star_id}'
                self.iw._marker = functools.partial(self.iw.dc.Text, text=label, fontsize=20, fontscale=False, color='green')
                self.iw.add_markers(Table(data=[[star['x']+20], [star['y']-20]], names=['x', 'y']),
                               marker_name=self._label_name)

            elif star['marker name'] == 'APASS comparison':
                label = f'C{star_id}'
                self.iw._marker = functools.partial(self.iw.dc.Text, text=label, fontsize=20, fontscale=False, color='red')
                self.iw.add_markers(Table(data=[[star['x']+20], [star['y']-20]], names=['x', 'y']),
                                    marker_name=self._label_name)

            elif star['marker name'] == 'VSX':
                label = f'V{star_id}'
                self.iw._marker = functools.partial(self.iw.dc.Text, text=label, fontsize=20, fontscale=False, color='blue')
                self.iw.add_markers(Table(data=[[star['x']+20], [star['y']-20]], names=['x', 'y']),
                                    marker_name=self._label_name)
            else:
                label = f'U{star_id}'
                print(f"Unrecognized marker name: {star['marker name']}")
            plot_names.append(label)

    def remove_labels(self):
        try:
            self.iw.remove_markers(marker_name=self._label_name)
        except AttributeError:
            self.iw.remove_markers_by_name(marker_name=self._label_name)

    def show_circle(self,
                    radius=2.5 * u.arcmin,
                    pixel_scale=0.56 * u.arcsec / u.pixel):
        radius_pixels = np.round((radius / pixel_scale).to(u.pixel).value,
                                 decimals=0)
        self.iw.marker = {'color': 'yellow',
                          'radius': radius_pixels,
                          'type': 'circle'}
        self.iw.add_markers(Table(data=[[self.target_coord]], names=['coords']),
                            skycoord_colname='coords',
                            use_skycoord=True, marker_name=self._circle_name)

    def remove_circle(self):
        try:
            self.iw.remove_markers(marker_name=self._circle_name)
        except AttributeError:
            self.iw.remove_markers_by_name(marker_name=self._circle_name)



import numpy as np

import pytest

from astropy.coordinates import SkyCoord
from astropy.table import Table
from astropy.time import Time
import astropy.units as u

from stellarphot.differential_photometry.aij_rel_fluxes import calc_aij_relative_flux


def _repeat(array, count):
    return np.concatenate([array for _ in range(count)])


def _raw_photometry_table():
    """
    Generate an input raw photometry table and expected flux ratios
    for use in tests.
    """

    n_times = 10
    n_stars = 4
    # How about ten times...
    times = Time('2018-06-25T01:00:00', format='isot', scale='utc')
    times = times + np.arange(n_times) * 30 * u.second
    times = times.value

    # and four stars
    star_ra = 250.0 * u.degree + np.arange(n_stars) * 10 * u.arcmin
    star_dec = np.array([45.0] * n_stars) * u.degree
    fluxes = np.array([10000., 20000, 30000, 40000])
    errors = np.sqrt(fluxes) + 50
    star_ids = np.arange(1, 5, dtype='int')

    # Stars 2, 3 and 4 will be the comparison stars
    comp_stars = np.array([0, 1, 1, 1])
    expected_comp_fluxes = np.sum(fluxes[1:])

    comp_flux_offset = - comp_stars * fluxes
    expected_flux_ratios = fluxes / (expected_comp_fluxes + comp_flux_offset)

    comp_error_total = np.sqrt((errors[1:]**2).sum())

    expected_flux_error = (fluxes / expected_comp_fluxes *
                           np.sqrt(errors**2 / fluxes**2 +
                                   comp_error_total**2 / expected_comp_fluxes**2))

    raw_table = Table(data=[np.sort(_repeat(times, n_stars)), _repeat(star_ra,  n_times),
                            _repeat(star_dec, n_times), _repeat(fluxes, n_times),
                            _repeat(errors, n_times),
                            _repeat(star_ids, n_times)],
                      names=['date-obs', 'RA', 'Dec', 'aperture_net_flux',
                             'noise-aij', 'star_id'])

    return expected_flux_ratios, expected_flux_error, raw_table, raw_table[1:4]


@pytest.mark.parametrize('comp_ra_dec_have_units', [True, False])
@pytest.mark.parametrize('star_ra_dec_have_units', [True, False])
@pytest.mark.parametrize('in_place', [True, False])
def test_relative_flux_calculation(in_place,
                                   star_ra_dec_have_units,
                                   comp_ra_dec_have_units):
    expected_flux, expected_error, input_table, comp_star = _raw_photometry_table()

    # Try doing it all at once
    n_times = len(np.unique(input_table['date-obs']))
    all_expected_flux = _repeat(expected_flux, n_times)
    all_expected_error = _repeat(expected_error, n_times)

    if not star_ra_dec_have_units:
        input_table['RA'] = input_table['RA'].data
        input_table['Dec'] = input_table['Dec'].data

    if not comp_ra_dec_have_units:
        comp_star['RA'] = comp_star['RA'].data
        comp_star['Dec'] = comp_star['Dec'].data

    output_table = calc_aij_relative_flux(input_table, comp_star,
                                          in_place=in_place)
    output_flux = output_table['relative_flux']
    output_error = output_table['relative_flux_error']
    print(all_expected_flux - output_flux)
    np.testing.assert_allclose(output_flux, all_expected_flux)
    np.testing.assert_allclose(output_error, all_expected_error)
    if in_place:
        assert 'relative_flux' in input_table.colnames
    else:
        assert 'relative_flux' not in input_table.colnames


@pytest.mark.parametrize('bad_thing', ['RA', 'NaN'])
def test_bad_comp_star(bad_thing):
    expected_flux, expected_error, input_table, comp_star = \
        _raw_photometry_table()
    # We'll do modify the "bad" property for the last star in the last
    # image.

    # First, let's sort so the row we want to modify is the last one
    input_table.sort(['date-obs', 'star_id'])

    # Force a copy of this row so we have access to the original values
    last_one = Table(input_table[-1])

    if bad_thing == 'RA':
        # "Jiggle" one of the stars by moving it by a few arcsec in one image.
        coord_inp = SkyCoord(ra=last_one['RA'], dec=last_one['Dec'],
                             unit=u.degree)
        coord_bad_ra = coord_inp.ra + 3 * u.arcsecond
        input_table['RA'][-1] = coord_bad_ra.degree
    elif bad_thing == 'NaN':
        input_table['aperture_net_flux'][-1] = np.nan

    output_table = calc_aij_relative_flux(input_table, comp_star,
                                          in_place=False)

    old_total_flux = comp_star['aperture_net_flux'].sum()
    new_flux = old_total_flux - last_one['aperture_net_flux']
    # This works for target stars, i.e. those never in comparison set
    new_expected_flux = old_total_flux / new_flux * expected_flux

    # Oh wow, this is terrible....
    # Need to manually calculate for the only two that are still in comparison
    new_expected_flux[1] = (comp_star['aperture_net_flux'][0] /
                            comp_star['aperture_net_flux'][1])
    new_expected_flux[2] = (comp_star['aperture_net_flux'][1] /
                            comp_star['aperture_net_flux'][0])

    new_expected_flux[3] = expected_flux[3]
    if bad_thing == 'NaN':
        new_expected_flux[3] = np.nan

    np.testing.assert_allclose(new_expected_flux, output_table['relative_flux'][-4:])



from astropy.wcs import WCS


def make_wcs():
    wcs = WCS(naxis=2)
    # Numbering of pixels for crpix starts at 1....
    wcs.wcs.crpix = [5., 5.]
    wcs.wcs.cdelt = [1, 1]
    wcs.wcs.crval = [10, 5]
    wcs.wcs.ctype = ["RA---TAN", "DEC--TAN"]
    return wcs



from pathlib import Path

from ccdproc import ImageFileCollection
import ipywidgets as ipw
from ipyfilechooser import FileChooser


class PhotometrySettings:
    def __init__(self):
        self._image_dir = FileChooser(title="Choose folder with images", show_only_dirs=True)
        self._aperture_file_loc = FileChooser(title='Choose aperture location file')
        self._aperture_settings_loc = FileChooser(title='Choose file with aperture settings')
        self._object_name = ipw.Dropdown(description='Choose object', style=dict(description_width='initial'))

        self._image_dir.register_callback(self._update_ifc)
        self._aperture_settings_loc.register_callback(self._update_aperture_rad)
        self.ifc = None
        self._box = ipw.VBox()
        self._box.children = [self._image_dir, self._aperture_file_loc, self._aperture_settings_loc, self._object_name]

    @property
    def box(self):
        return self._box

    @property
    def image_folder(self):
        return self._image_dir.selected
    
    @property
    def aperture_locations(self):
        return self._aperture_file_loc.selected
    
    @property
    def aperture_radius(self):
        return self._aperture_radius

    @property
    def object_name(self):
        return self._object_name.value

    def _update_ifc(self, change):
        self.ifc = ImageFileCollection(change.selected)
        self._update_object_list(change)

    def _update_object_list(self, change):
        if self.ifc.summary:
            self._object_name.options = sorted(set(self.ifc.summary['object']))
        else:
            self._object_name.options = []
    
    def _update_aperture_rad(self, locator):
        with open(locator.selected) as f:
            stuff = f.read()
        
        self._aperture_radius = int(stuff.split(',')[0])



from collections import namedtuple

import numpy as np

from astropy.modeling import models, fitting
from astropy.stats import sigma_clip
from astropy.table import MaskedColumn
from astropy.coordinates import SkyCoord, match_coordinates_sky
from astropy import units as u

__all__ = [
    'filter_transform',
    'calculate_transform_coefficients',
    'transform_magnitudes'
]


def filter_transform(mag_data, output_filter,
                     g=None, r=None, i=None,
                     transform=None):
    '''
    Transform SDSS magnitudes to BVRI using either the transforms from
    Jester et al or Ivezic et al.

    Parameters
    ----------

    mag_data : astropy.table.Table
        Table containing ``g``, ``r`` and ``i`` magnitudes (or at least)
        those required to transform to the desired output filter.
    output_filter : 'B', 'V', 'R' or 'I'
        Filter for which magnitude should be calculated. Note that
        *case matters* here.
    g, r, i : str
        Name of column in table for that magnitude.
    transform : 'jester' or 'ivezic'
        Transform equations to use.
    Description: This function impliments the transforms in
        'A Comparison of SDSS Standard

    Returns
    -------

    astropy.table.Column
        Output transformed magnitudes as a table column

    Notes
    -----

    The transforms implemented in this function are taken from:

    Jester, et al, *The Sloan Digital Sky Survey View of the Palomar-Green Bright Quasar Survey*, AJ 130, p. 873 (2005)
    http://iopscience.iop.org/article/10.1086/432466/meta

    Ivezić et al, *A Comparison of SDSS Standard Star Catalog for Stripe 82 with Stetson's Photometric Standards*,
    The Future Of Photometric, Spectrophotometric And Polarimetric Standardization, ASP Conference Series 364, p. 165 (2007)
    http://aspbooks.org/custom/publications/paper/364-0165.html

    '''
    supported_transforms = ['jester', 'ivezic']
    if transform not in supported_transforms:
        raise ValueError('Transform {} is not known. Must be one of '
                         '{}'.format(transform, supported_transforms))
    transform_ivezic = {
        'B': [0.2628, -0.7952, 1.0544, 0.0268],
        'V': [0.0688, -0.2056, -0.3838, -0.0534],
        'R': [-0.0107, 0.0050, -0.2689, -0.1540],
        'I': [-0.0307, 0.1163, -0.3341, -0.3584]
    }
    base_mag_ivezic = {
        'B': g,
        'V': g,
        'R': r,
        'I': i
    }
    # For jester, using the transform for "all stars with Rc-Ic < 1.15"
    # from
    # http://www.sdss3.org/dr8/algorithms/sdssUBVRITransform.php#Jester2005
    jester_transforms = {
        'B': [1.39, -0.39, 0, 0.21],
        'V': [0.41, 0.59, 0, -0.01],
        'R': [0.41, -0.5, 1.09, -0.23],
        'I': [0.41, -1.5, 2.09, -0.44]
    }

    if output_filter not in base_mag_ivezic.keys():
        raise ValueError('the desired filter must be a string R B V or I')

    if transform == 'ivezic':
        if output_filter == 'R' or output_filter == 'I':
            # This will throw a KeyError if the column is missing
            c = mag_data[r] - mag_data[i]

        if output_filter == 'B' or output_filter == 'V':
            # This will throw a KeyError if the column is missing
            c = mag_data[g] - mag_data[r]

        transform_poly = np.poly1d(transform_ivezic[output_filter])
        out_mag = transform_poly(c) + \
            mag_data[base_mag_ivezic[output_filter]]
        # poly1d  ignores masks. Add masks back in here if necessary.
        try:
            input_mask = c.mask
        except AttributeError:
            pass
        else:
            out_mag = np.ma.array(out_mag, mask=input_mask)
    elif transform == 'jester':
        coeff = jester_transforms[output_filter]
        out_mag = (coeff[0] * mag_data[g] + coeff[1] * mag_data[r] +
                   coeff[2] * mag_data[i] + coeff[3])

    out_mag.name = '{}_mag'.format(output_filter)
    out_mag.description = ('{}-band magnitude transformed '
                           'from gri'.format(output_filter))
    return out_mag


def calculate_transform_coefficients(input_mag, catalog_mag, color,
                                     input_mag_error=None,
                                     catalog_mag_error=None,
                                     faintest_mag=None,
                                     order=1,
                                     sigma=2.0,
                                     gain=None,
                                     ):
    """
    Calculate linear transform coefficients from input magnitudes to catalog
    magnitudes.

    Parameters
    ----------

    input_mag : numpy array or astropy Table column
        Input magnitudes; for example, instrumental magnitudes.
    catalog_mag : numpy array or astropy Table column
        Catalog (or reference) magnitudes; the magnitudes to which the
        input_mag will eventually be transformed.
    color : numpy array or astropy Table column
        Colors to use in determining transform coefficients.
    input_mag_error : numpy array or astropy Table column, optional
        Error in input magnitudes. Default is zero.
    catalog_mag_error : numpy array or astropy Table column, optional
        Error in catalog magnitudes. Default is zero.
    faintest_mag_for_transform : float, optional
        If this is not ``None``, the magnitude of the faintest catalog stars
        to use in computing transform coefficients.
    order : int, optional
        Order of the polynomial fit to use in correcting for color.
    sigma : float, optional
        Value of sigma to use to reject outliers while fitting using
        sigma clipping.
    gain : float, optional
        If not ``None``, adjust the instrumental magnitude by
        -2.5 * log10(gain), i.e. gain correct the magnitude.
    verbose : bool, optional
        If ``True``, print some diagnostic information.
    extended_output : bool, optional
        If ``True``, return additional information.

    Returns
    -------

    filtered_data : `~numpy.ma.core.MaskedArray`
        The data, with the mask set ``True`` for the data that was *omitted*
        from the fit.

    model : `astropy.modeling.FittableModel`
        Entries in the model are the coefficients in the fit made to the
        data. Since the model is always a polynomial, these are terms in
        a polynomial in the order of ascending power. In other words, the
        coefficient ``ci`` is the coefficient of the term ``x**i``.

    If ``extended_output=True``, then also return:

    fit_input : tuple
        A tuple of color, magnitude for only the stars brighter than
        ``faintest_mag_for_transform``. These are input to the sigma-clipping
        fitter.

    used_in_fit : tuple
        A tuple of color, magnitude for only the stars brighter than
        ``faintest_mag_for_transform`` that were not sigma-cliped out.

    Notes
    -----

    This function has some pretty serious limitations right now:

    + Errors in the independent variable are ignored.
    + Outliers are rejected using a modified loss function (Huber loss) that
      cannot be modified.
    + No errors are estimated in the calculated transformation coefficients.

    And there is all the stuff that is not listed here...
    """

    if input_mag_error is None:
        input_mag_error = np.zeros_like(input_mag)
    if catalog_mag_error is None:
        catalog_mag_error = np.zeros_like(catalog_mag)

    if gain is None:
        gain = 1.0

    # Independent variable is the color, dependent variable is the
    # difference between the measured (input) magnitude and the catalog
    # magnitude.

    mag_diff = catalog_mag - (input_mag - 2.5 * np.log10(gain))

    # The error is the errors of those combined in quadrature.
    combined_error = np.sqrt(input_mag_error**2 + catalog_mag_error**2)

    # If both errors are zero then the error is omitted.
    if (combined_error == 0).all():
        dy = None
    else:
        dy = combined_error

    g_init = models.Polynomial1D(order)
    fit = fitting.LinearLSQFitter()
    or_fit = fitting.FittingWithOutlierRemoval(fit, sigma_clip,
                                               niter=2, sigma=sigma)

    if faintest_mag is not None:
        bright = catalog_mag < faintest_mag
        try:
            bright = bright.filled(False)
        except AttributeError:
            # Might not have had a masked array...
            pass
    else:
        bright = np.ones_like(mag_diff, dtype='bool')

    bright_index = np.nonzero(bright)

    # get fitted model and filtered data
    or_fitted_model, filtered_data_mask = or_fit(g_init,
                                                 color[bright],
                                                 mag_diff[bright])

    # Restore the filtered_data to the same size as the input
    # magnitudes. Unmasked values were included in the fit,
    # masked were not, either because they were too faint
    # or because they were sigma clipped out.
    restored_mask = np.zeros_like(mag_diff, dtype='bool')
    restored_mask[bright_index] = filtered_data_mask
    restored_mask[~bright] = True

    restored_filtered = MaskedColumn(mag_diff.copy())
    restored_filtered.mask = restored_mask

    return (restored_filtered, or_fitted_model)


def transform_magnitudes(input_mags, catalog,
                         transform_catalog,
                         input_mag_colum='mag_inst_r',
                         catalog_mag_column='r_mag',
                         catalog_color_column='B-V',
                         faintest_mag_for_transform=14,
                         sigma=2,
                         order=1,
                         gain=None):
    """
    Calculate catalog magnitudes and transform coefficients
    from instrumental magnitudes.

    Parameters
    ----------

    input_mags : astropy Table
        Table which contains a column with instrumental magnitudes, i.e.
        -2.5 * log10(net_counts / exposure_time).

    catalog : astropy Table
        Table containing reference catalog of magnitudes and colors.


    transform_catalog : astropy Table
        Table containing the reference catalog of magnitudes and colors
        to use in determining the transform coefficients. Can be the
        same table as ``catalog`` if desired.

    input_mag_column : str, optional
        Name of the column in ``input_mags`` with the magnitudes to be
        transformed.

    catalog_mag_column : str, optional
        Name of the column in ``catalog`` with the reference magnitude.

    catalog_color_column : str, optional
        Name of the column in ``catalog`` with color for each star in the
        catalog.

    faintest_mag_for_transform : float, optional
        If this is not ``None``, the magnitude of the faintest catalog stars
        to use in computing transform coefficients.

    sigma : float, optional
        Number of standard deviations to use in rejecting outliers when fitting
        using sigma clipping.

    order : int, optional
        Order of the polynomial to use in fitting magnitude difference/color
        relationship.

    gain : float, optional
        If not ``None``, adjust the instrumental magnitude by
        -2.5 * log10(gain), i.e. gain correct the magnitude.
    """
    catalog_all_coords = SkyCoord(catalog['RAJ2000'],
                                  catalog['DEJ2000'],
                                  unit='deg')

    transform_catalog_coords = SkyCoord(transform_catalog['RAJ2000'],
                                        transform_catalog['DEJ2000'],
                                        unit='deg')
    input_coords = SkyCoord(input_mags['RA'], input_mags['Dec'])

    transform_catalog_index, d2d, _ = \
        match_coordinates_sky(input_coords, transform_catalog_coords)

    # create a boolean of all of the matches that have a discrepancy of less
    # than 5 arcseconds
    good_match_for_transform = d2d < 2 * u.arcsecond

    catalog_index, d2d, _ = match_coordinates_sky(input_coords,
                                                  catalog_all_coords)

    good_match_all = d2d < 5 * u.arcsecond

    catalog_all_indexes = catalog_index[good_match_all]

    input_match_mags = input_mags[input_mag_colum][good_match_for_transform]

    catalog_match_indexes = transform_catalog_index[good_match_for_transform]

    catalog_match_mags = \
        transform_catalog[catalog_mag_column][catalog_match_indexes]
    catalog_match_color = \
        transform_catalog[catalog_color_column][catalog_match_indexes]

    good_mags = ~np.isnan(input_match_mags)

    input_match_mags = input_match_mags[good_mags]
    catalog_match_mags = catalog_match_mags[good_mags]
    catalog_match_color = catalog_match_color[good_mags]

    try:
        matched_data, transforms = \
            calculate_transform_coefficients(input_match_mags,
                                             catalog_match_mags,
                                             catalog_match_color,
                                             sigma=sigma,
                                             faintest_mag=faintest_mag_for_transform,
                                             order=order,
                                             gain=gain)
    except np.linalg.LinAlgError as e:
        print('Danger! LinAlgError: {}'.format(str(e)))
        Transform = namedtuple('Transform', ['parameters'])
        transforms = Transform(parameters=(np.nan,) * (order + 1))

    our_cat_mags = (input_mags[input_mag_colum][good_match_all] +
                    transforms(catalog[catalog_color_column][catalog_all_indexes]))

    return our_cat_mags, good_match_all, transforms



# Licensed under a 3-clause BSD style license - see LICENSE.rst

__all__ = ['__version__']

try:
    from .version import version as __version__
except ImportError:
    __version__ = ''

import os
from warnings import warn
from astropy.config.configuration import (
    update_default_config,
    ConfigurationDefaultMissingError,
    ConfigurationDefaultMissingWarning)

# Create the test function for self test
from astropy.tests.runner import TestRunner
test = TestRunner.make_test_runner_in(os.path.dirname(__file__))
test.__test__ = False
__all__ += ['test']

# add these here so we only need to cleanup the namespace at the end
config_dir = None

if not os.environ.get('ASTROPY_SKIP_CONFIG_UPDATE', False):
    config_dir = os.path.dirname(__file__)
    config_template = os.path.join(config_dir, __package__ + ".cfg")
    if os.path.isfile(config_template):
        try:
            update_default_config(
                __package__, config_dir, version=__version__)
        except TypeError as orig_error:
            try:
                update_default_config(__package__, config_dir)
            except ConfigurationDefaultMissingError as e:
                wmsg = (e.args[0] +
                        " Cannot install default profile. If you are "
                        "importing from source, this is expected.")
                warn(ConfigurationDefaultMissingWarning(wmsg))
                del e
            except Exception:
                raise orig_error



import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from astropy.stats import mad_std
from astropy.time import Time

from gatspy.periodic import LombScargleFast


__all__ = ['plot_magnitudes', 'multi_night']


def plot_magnitudes(mags=None, errors=None, times=None,
                    source=None, night=None, ref_mag=0,
                    alpha=0.25, y_range=None):
    """
    Plot one night of magnitude data for one source, overlaying a rolling
    mean and indication of mean/deviation.
    """
    mean = np.nanmean(mags)
    std = np.nanstd(mags)

    working_times = times
    plt.errorbar(working_times, mags, yerr=errors, fmt='o', alpha=alpha,
                 label='night: {}'.format(night))
    plt.xlim(working_times.min(), working_times.max())
    plt.plot(plt.xlim(), [mean, mean], 'k--', )
    plt.plot(plt.xlim(), [mean + std, mean + std], 'k:')
    plt.plot(plt.xlim(), [mean - std, mean - std], 'k:')
    plt.plot(pd.rolling_mean(working_times, 20, center=True),
             pd.rolling_mean(mags, 20, center=True),
             color='gray', linewidth=3)

    if y_range:
        plt.ylim(y_range)
    else:
        # Make sure plot range is at least 0.1 mag...
        min_range = 0.1
        ylim = plt.ylim()
        if ylim[1] - ylim[0] < min_range:
            plt.ylim(mean - min_range / 2, mean + min_range / 2)

    ylim = plt.ylim()
    # Reverse vertical axis so brighter is higher
    plt.ylim((ylim[1], ylim[0]))

    # # Add marker indicating brightness of star.
    # size = 1000./(mean - ref_mag + 0.1)**2
    # plt.scatter([0.8*(plt.xlim()[1]-plt.xlim()[0]) + plt.xlim()[0]],
    #             [0.8*(plt.ylim()[1] - plt.ylim()[0]) + plt.ylim()[0]],
    #             c='red', marker='o', s=size)

    # plt.legend()
    calendar_date = Time(night, format='jd', out_subfmt='date')
    calendar_date.format = 'iso'

    plt.title('night {}'.format(calendar_date))
    plt.xlabel('time (days)')
    return mean, std


def multi_night(sources, unique_nights, night,
                brightest_mag, mags, mag_err,
                uniform_ylim=True):
    """
    Plot magnitude vs time data for several sources over several nights
    """
    number_of_nights = len(unique_nights)

    for source in sources:
        f = plt.figure(figsize=(5 * number_of_nights, 5))

        night_means = []
        night_stds = []
        night_bins = []
        source_mags = mags[source.id - 1]
        if uniform_ylim:
            # Use median to handle outliers.
            source_median = np.median(source_mags[np.isfinite(source_mags)])
            # Use median absolute deviation to get measure of scatter.
            # Helps avoid extremely points.
            source_variation = \
                3 * mad_std(source_mags[np.isfinite(source_mags)])

            # Ensure y range will be at least 0.2 magnitudes
            if source_variation < 0.1:
                half_range = 0.1
            else:
                half_range = source_variation

            y_range = (source_median - half_range, source_median + half_range)
        else:
            # Empty if this option wasn't chosen so that automatic limits
            # will be used.
            y_range = []

        last_axis = None
        for i, this_night in enumerate(unique_nights):
            last_axis = plt.subplot(1, number_of_nights + 1, i + 1,
                                    sharey=last_axis)
            night_mask = (night == this_night)
            night_mean, night_std = \
                plot_magnitudes(mags=mags[source.id - 1][night_mask],
                                errors=mag_err[source.id - 1][night_mask],
                                times=source.bjd_tdb[night_mask],
                                source=source.id,
                                night=this_night,
                                ref_mag=brightest_mag,
                                y_range=y_range)
            night_means.append(night_mean)
            night_stds.append(night_std)
            night_bins.append(this_night)

        plt.subplot(1, number_of_nights + 1, number_of_nights + 1)

        if uniform_ylim:
            f.subplots_adjust(wspace=0)
            plt.setp([a.get_yticklabels() for a in f.axes[1:]], visible=False)

        # Plot indicators of variation, and information about this source.
        # For simplicity, make the x and y range of this plot be 0 to 1.
        x = np.array([0., 1])
        y = x

        # Add invisible line to make plot.
        plt.plot(x, y, alpha=0, label='source {}'.format(source.id))
        night_means = np.array(night_means)

        # Plot bar proportional to Lomb-Scargle power.
        bad_mags = (np.isnan(mags[source.id - 1]) |
                    np.isinf(mags[source.id - 1]))
        bad_errs = (np.isnan(mag_err[source.id - 1]) |
                    np.isinf(mag_err[source.id - 1]))
        bads = bad_mags | bad_errs
        good_mags = ~bads
        model = LombScargleFast().fit(source.bjd_tdb[good_mags],
                                      mags[source.id - 1][good_mags],
                                      mag_err[source.id - 1][good_mags])
        periods, power = model.periodogram_auto(nyquist_factor=100,
                                                oversampling=3)
        max_pow = power.max()

        # print(source, max_pow)
        if max_pow > 0.5:
            color = 'green'
        elif max_pow > 0.4:
            color = 'cyan'
        else:
            color = 'gray'

        bar_x = 0.25
        plt.plot([bar_x, bar_x], [0, max_pow],
                 color=color, linewidth=10, label='LS power')

        plt.legend()

        # Add dot for magnitude of star.
        size = 10000. / np.abs(10**((source_median - brightest_mag) / 2.5))
        plt.scatter([0.8], [0.2], c='red', marker='o', s=size)
        plt.ylim(0, 1)



import numpy as np

from astropy.coordinates import SkyCoord
from astropy.table import Table
import astropy.units as u

__all__ = ['add_in_quadrature', 'calc_aij_relative_flux']

def add_in_quadrature(array):
    """
    Add an array of numbers in quadrature.
    """
    return np.sqrt((array**2).sum())


def calc_aij_relative_flux(star_data, comp_stars,
                           in_place=True, coord_column=None, 
                           star_id_column='star_id',
                           flux_column_name='aperture_net_flux'):
    """
    Calculate AstroImageJ-style flux ratios.

    Parameters
    ----------

    star_data : '~astropy.table.Table'
        Table of star data from one or more images.

    comp_stars : '~astropy.table.Table'
        Table of comparison stars in the field. Must contain a column
        called ``RA`` and a column called ``Dec``.
        NOTE that not all
        of the comparison stars will necessarily be used. Stars in
        this table are excluded from the comparison set if, in any
        of the `star_data` for that comparison, the net counts are
        ``NaN`` or if the angular distance between the position in
        the `star_data` and the position in the `comp_stars` table
        is too large.

    in_place : ``bool``, optional
        If ``True``, add new columns to input table. Otherwise, return
        new table with those columns added.

    coord_column : ``str``, optional
        If provided, use this column to match comparison stars to coordinates.
        If not provided, the coordinates are generated with SkyCoord.

    flux_column_name : ``str``, optional
        If provided, use this column to find flux.
        If not provided, the column 'aperture_net_flux' is used.

    star_id_column : ``str``, optional
        Name of the column that provides a unique identifier for each
        comparison star.

    Returns
    -------

    `astropy.table.Table` or None
        The return type depends on the value of ``in_place``. If it is
        ``False``, then the new columns are returned as a separate table,
        otherwise the columns are simply added to the input table.
    """

    # Match comparison star list to instrumental magnitude information
    if star_data['RA'].unit is None:
        unit = 'degree'
    else:
        # Pulled this from the source code -- None is ok but need
        # to match the number of coordinates.
        unit = [None, None]

    star_data_coords = SkyCoord(ra=star_data['RA'], dec=star_data['Dec'],
                                unit=unit)

    if coord_column is not None:
        comp_coords = comp_stars[coord_column]
    else:
        if comp_stars['RA'].unit is None:
            unit = 'degree'
        else:
            # Pulled this from the source code -- None is ok but need
            # to match the number of coordinates.
            unit = [None, None]
        comp_coords = SkyCoord(ra=comp_stars['RA'], dec=comp_stars['Dec'],
                               unit=unit)

    # Check for matches of stars in star data to the stars in comp_stars
    # and eliminate as comps any stars for which the separation is bigger
    # than 1.2 arcsec in any of the frames.
    index, d2d, _ = star_data_coords.match_to_catalog_sky(comp_coords)

    # Not sure this is really close enough for a good match...
    good = d2d < 1.2 * u.arcsec

    check_for_bad = Table(data=[star_data[star_id_column].data, good],
                          names=['star_id', 'good'])
    check_for_bad = check_for_bad.group_by('star_id')
    is_all_good = check_for_bad.groups.aggregate(np.all)

    bad_comps = set(is_all_good['star_id'][~is_all_good['good']])
    print(f'{bad_comps=}')

    # Check whether any of the comp stars have NaN values and,
    # if they do, exclude them from the comp set.
    check_for_nan = Table(data=[star_data[star_id_column].data,
                                star_data[flux_column_name].data],
                          names=['star_id', 'net_counts'])
    check_for_nan = check_for_nan.group_by('star_id')
    check_for_nan['good'] = ~np.isnan(check_for_nan['net_counts'])
    is_all_good = check_for_nan.groups.aggregate(np.all)

    bad_comps = bad_comps | set(is_all_good['star_id'][~is_all_good['good']])

    print(f'{bad_comps=}')
    for comp in bad_comps:
        this_comp = star_data[star_id_column] == comp
        good[this_comp] = False

    error_column_name = 'noise-aij'
    # Calculate comp star counts for each time

    # Make a small table with just counts, errors and time for all of the comparison
    # stars.

    comp_fluxes = star_data['date-obs', flux_column_name, error_column_name][good]
    # print(np.isnan(comp_fluxes[flux_column_name]).sum(),
    #       np.isnan(comp_fluxes[error_column_name]).sum())
    # print(star_data[good][flux_column_name][np.isnan(comp_fluxes[flux_column_name])])

    # Check whether any of the columns are masked, but with no masked values,
    # and convert to regular column...eventually

    comp_fluxes = comp_fluxes.group_by('date-obs')
    comp_totals = comp_fluxes.groups.aggregate(np.sum)[flux_column_name]
    comp_num_stars = comp_fluxes.groups.aggregate(np.count_nonzero)[flux_column_name]
    comp_errors = comp_fluxes.groups.aggregate(add_in_quadrature)[error_column_name]

    comp_total_vector = np.ones_like(star_data[flux_column_name])
    comp_error_vector = np.ones_like(star_data[flux_column_name])

    if len(set(comp_num_stars)) > 1:
        raise RuntimeError('Different number of stars in comparison sets')

    # Calculate relative flux for every star

    # Have to remove the flux of the star if the star is a comparison
    # star.
    is_comp = np.zeros_like(star_data[flux_column_name])
    is_comp[good] = 1
    flux_offset = -star_data[flux_column_name] * is_comp

    # This seems a little hacky; there must be a better way
    for date_obs, comp_total, comp_error in zip(comp_fluxes.groups.keys,
                                                comp_totals, comp_errors):
        this_time = star_data['date-obs'] == date_obs[0]
        comp_total_vector[this_time] *= comp_total
        comp_error_vector[this_time] = comp_error

    relative_flux = star_data[flux_column_name] / (comp_total_vector + flux_offset)
    relative_flux = relative_flux.flatten()

    rel_flux_error = (star_data[flux_column_name] / comp_total_vector *
                      np.sqrt((star_data[error_column_name] / star_data[flux_column_name])**2 +
                              (comp_error_vector / comp_total_vector)**2
                              )
                     )

    # Add these columns to table
    if not in_place:
        star_data = star_data.copy()

    star_data['relative_flux'] = relative_flux
    star_data['relative_flux_error'] = rel_flux_error
    star_data['relative_flux_snr'] = relative_flux / rel_flux_error

    # AIJ records the total comparison counts even though that total is used
    # only for the targets, not the comparison.
    star_data['comparison counts'] = comp_total_vector  # + flux_offset
    star_data['comparison error'] = comp_error_vector

    return star_data



def convert_pixel_wcs(ccd_image, lon_or_ra,
                      lat_or_dec, is_pix=True, origin=0):
    """
    Takes either pixel or World Coordinate System (RA/Dec)
    coordinates and converts to the other type.

    Parameters
    ----------
    ccd_image: CCDData object
        Image which has the coordinates to convert.

    lon_or_ra: `numpy.ndarray' (or float)
        An array of coordinates (or a single coordinate).
        Represents the lon/ra axis, depending on what is
        passed in.

    lat_or_dec: `numpy.ndarray' (or float)
        An array of coordinates (or a single coordinate).
        Represents the lat/dec axis, depending on what is
        passed in.

    is_pix: bool
        Used to determine what coordinates are being inputted.
        Assumes pixel coordinates by default.

    origin: int
        The coordinate in the upper left corner of the image.
        Generally, this is 1 in FITS and Fortran standards,
        or 0 in Numpy and C standards. Defaults to 0.

    Returns
    -------
    Returns two `numpy.ndarray' objects, the first being lon/RA,
    and the second lat/Dec.

    """
    if is_pix:
        return ccd_image.wcs.all_pix2world(lon_or_ra, lat_or_dec, origin)
    else:
        return ccd_image.wcs.all_world2pix(lon_or_ra, lat_or_dec, origin)



from astropy import units as u

__all__ = ['Camera']


class Camera:
    """docstring for Camera"""
    def __init__(self, gain=1.0 * u.electron / u.adu,
                 read_noise=1.0 * u.electron,
                 dark_current=0.01 * u.electron / u.second):
        super().__init__()
        self.gain = gain
        self.read_noise = read_noise
        self.dark_current = dark_current



import re
from pathlib import Path
import json

import numpy as np
import ipywidgets as ipw
from traitlets import observe, Bool

from astroquery.mast import Catalogs
from astropy.utils.data import get_pkg_data_filename


template_types = ['known', 'candidate']
template_json = {}
to_fill = {}

for template in template_types:
    template_name = get_pkg_data_filename('data/tic-template-for-exotic-'
                                          f'{template}.json')
    with open(template_name) as f:
        template_json[template] = json.load(f)

    template_name = get_pkg_data_filename(f'data/exotic-to-mod-{template}.json')
    with open(Path(template_name)) as f:
        to_fill[template] = json.load(f)

exotic_arguments = dict(
    known=['--nasaexoarch', '--pre'],
    candidate=['--override', '--pre']
)

join_char = "😬"

exotic_tic = {
    "Star Effective Temperature (K)": "Teff",
    "Star Effective Temperature (+) Uncertainty": "epos_Teff",
    "Star Effective Temperature (-) Uncertainty": "eneg_Teff",
    "Star Surface Gravity (log(g))": "logg",
    "Star Surface Gravity (+) Uncertainty": "epos_logg",
    "Star Surface Gravity (-) Uncertainty": "eneg_logg",
    "Host Star Name": "UCAC",
    "Star Metallicity ([FE/H])": "MH",
    "Star Metallicity (+) Uncertainty": "e_MH",
    "Star Metallicity (-) Uncertainty": "e_MH"
}


class MyValid(ipw.Button):
    """
    A more compact indicator of valid entries.
    """
    value = Bool(False, help="Bool value").tag(sync=True)

    def __init__(self, **kwd):
        super().__init__(**kwd)
        self.layout.width = '40px'
        self._set_properties(None)

    @observe('value')
    def _set_properties(self, change):
        if self.value:
            self.style.button_color = 'green'
            self.icon = 'check'
        else:
            self.style.button_color = 'red'
            self.icon = 'times'


def get_tic_info(TIC_ID):
    catalog_data = Catalogs.query_criteria(catalog="Tic", ID=TIC_ID)
    return catalog_data


def make_checker(indicator_widget, value_widget):
    def check_name(change):
        # Valid TIC number is 9 digits
        ticced = re.compile(r'TIC \d{9,10}$')
        owner = change['owner']
        is_tic = ticced.match(change['new'])
        if is_tic:
            if indicator_widget is not None:
                indicator_widget.value = True
            owner.disabled = True
            tic_info = get_tic_info(change['new'][-9:])
            if not tic_info:
                indicator_widget.value = False
                indicator_widget.tooltip = "Not a valid TIC number"
            else:
                populate_boxes(tic_info, value_widget)
            owner.disabled = False
        else:
            owner.disabled = False
            if indicator_widget is not None:
                indicator_widget.value = False
                indicator_widget.tooltip = 'TIC numbers have 9 digits'

    return check_name


def validate_exposure_time(indicator_widget, value_widget):
    def check_exposure(change):
        # Valid Exposure time is greater than zero
        if change['new'] > 0:
            if indicator_widget is not None:
                indicator_widget.value = True
        else:
            if indicator_widget is not None:
                indicator_widget.value = False
    return check_exposure


def populate_boxes(tic_info, value_widget):
    """
    Set the appropriate widget values given information pulled from the
    TIC.
    """
    for k, v in exotic_tic.items():
        exotic_key = join_char.join(["planetary_parameters", k])
        if k == "Host Star Name":
            value_widget['candidate'][exotic_key].value = \
                f'UCAC4 {tic_info[v][0]}'
        elif not np.isnan(tic_info[v][0]):
            value_widget['candidate'][exotic_key].value = tic_info[v][0]


validators = dict(known={}, candidate={})
validators['candidate']['Planet Name'] = make_checker
for k in validators:
    validators[k]['Exposure Time (s)'] = validate_exposure_time


def exotic_settings_widget(init_from_json=None):
    """
    Generate a widget to enter settings from sample.
    """

    # We rely on some global variables:
    global to_fill, template_types
    # This dictionary will contain all of the widgets
    widget_list = {}

    # Each widget has the same layout for its description and its value/input
    layout_description = ipw.Layout(width='70%')
    layout_input = ipw.Layout(width='30%')

    # Maintain a separate dict of just the value widgets
    value_widget = {}

    # For exotic there are two templates, one for known exoplanets and one for
    # candidate exoplanets
    for template in template_types:
        value_widget[template] = {}
        widget_list[template] = []
        for k in to_fill[template]:
            for k2, v in to_fill[template][k].items():
                if isinstance(v, str):
                    input_widget = ipw.Text(value=v)
                elif isinstance(v, float):
                    if v >= 0:
                        input_widget = ipw.FloatText(value=v)
                    else:
                        input_widget = ipw.FloatText(value=v)
                input_widget.layout = layout_input
                hb = ipw.HBox([ipw.HTML(value=k2, layout=layout_description),
                               input_widget])
                try:
                    validator = validators[template][k2]
                except KeyError:
                    pass
                else:
                    kids = list(hb.children)
                    indicator = MyValid(value=False)
                    kids.append(indicator)
                    input_widget.observe(validator(indicator, value_widget),
                                         names='value')

                    hb.children = kids

                val_key = join_char.join([k, k2])
                value_widget[template][val_key] = input_widget
                widget_list[template].append(hb)

    hb2 = {}
    for template in template_types:
        hb2[template] = ipw.HBox([ipw.VBox(widget_list[template][:16],
                                           layout=ipw.Layout(padding='10px')),
                                  ipw.VBox(widget_list[template][16:])])

    select_planet_type = ipw.ToggleButtons(
        description='Known or candidate exoplanet?',
        options=template_types,
        style={'description_width': 'initial'}
    )

    lookup_link_text = dict(known='https://exoplanetarchive.ipac.caltech.edu/',
                            candidate='https://exofop.ipac.caltech.edu/tess/')

    lookup_link_html = {}

    for k, v in lookup_link_text.items():
        lookup_link_html[k] = ipw.HTML(
            f'<h3>For some information about this '
            f'object: <a href="{v}" target="_blank">{v}</a></h3>'
        )

    input_container = ipw.VBox()

    whole_thing = ipw.VBox(children=[select_planet_type, input_container])
    whole_thing.planet_type = select_planet_type
    whole_thing.value_widget = value_widget
    pre_reduced_file = join_char.join(['optional_info', 'Pre-reduced File:'])
    whole_thing.data_file_widget = {
        'candidate': value_widget['candidate'][pre_reduced_file],
        'known': value_widget['known'][pre_reduced_file]
    }

    def observe_select(change):
        input_container.children = [lookup_link_html[select_planet_type.value],
                                    hb2[select_planet_type.value]]

    select_planet_type.observe(observe_select, names='value')
    observe_select(select_planet_type.value)

    return whole_thing


def set_values_from_json_file(widget, json_file):
    with open(json_file) as f:
        input_values = json.load(f)

    planet_type = widget.planet_type.value
    for k, widget in widget.value_widget[planet_type].items():
        k1, k2 = k.split(join_char)
        widget.value = input_values[k1][k2]


def get_values_from_widget(key, whole_thing):
    for k, widget in whole_thing.value_widget[key].items():
        k1, k2 = k.split(join_char)
        template_json[key][k1][k2] = widget.value
    return template_json[key]


def generate_json_file_name(key, whole_thing):
    get_values_from_widget(key, whole_thing)
    user_info = 'user_info'
    planet = 'planetary_parameters'
    filter_key = 'Filter Name (aavso.org/filters)'
    date = template_json[key][user_info]['Observation date']
    planet_name = template_json[key][planet]['Planet Name']
    filter_name = template_json[key][user_info][filter_key]
    name = f'{planet_name}-{date}-{filter_name}'
    return name.replace(' ', '_') + '.json'
